@@ -8,24 +8,36 @@
 # Perhaps at some point, it will be re-added,
 # though it doesn't sound too useful atm. :P
 
+import importlib
 import asyncio
 import socket
+import signal
 import os
 import re
-from collections import defaultdict
-from enum import IntEnum, unique
-from typing import AsyncGenerator, Optional, Union
+import time
+import gzip
+from enum import IntEnum
+from enum import unique
+from typing import Callable, Optional
+from typing import Coroutine
+from typing import Iterable
+from typing import Optional
+from typing import Union
 
-from .logging import log, Ansi
+from .logging import log
+from .logging import Ansi
+from .logging import AnsiRGB
 
 __all__ = (
-    # Information
+    # Informational
     'HTTPStatus',
     'Address',
 
-    # Asynchronous
-    'AsyncConnection',
-    'AsyncTCPServer'
+    # Functional
+    'Connection',
+    'RouteMap',
+    'Domain',
+    'Server'
 )
 
 _httpstatus_str = {
@@ -191,16 +203,18 @@ req_line_re = re.compile(
     r'HTTP/(?P<httpver>1\.0|1\.1|2\.0|3\.0)$'
 )
 
-class AsyncConnection:
+class Connection:
     __slots__ = (
         'client',
 
         # Request params
         'headers', 'body', 'cmd',
-        'path', 'httpver', 'args', 'files',
+        'path', 'httpver',
+
+        'args', 'multipart_args', 'files',
 
         # Response params
-        'resp_headers', 'multipart_args'
+        'resp_code', 'resp_headers',
     )
 
     # TODO: probably cut back on the defaultdicts
@@ -208,18 +222,20 @@ class AsyncConnection:
         self.client = client
 
         # Request params
-        self.headers: defaultdict[str, str] = defaultdict(lambda: None)
+        self.headers = {}
         self.body: Optional[bytearray] = None
-        self.cmd: Optional[str] = None
-        self.path: Optional[str] = None
-        self.httpver: Optional[float] = None
+        self.cmd = ''
+        self.path = ''
+        self.httpver = 0.0
 
-        self.args: defaultdict[str, str] = defaultdict(lambda: None)
-        self.files: defaultdict[str, bytes] = defaultdict(lambda: None)
+        self.args = {}
+        self.multipart_args = {}
+
+        self.files = {}
 
         # Response params
+        self.resp_code = 200
         self.resp_headers: list[str] = []
-        self.multipart_args: defaultdict[str, str] = defaultdict(lambda: None)
 
     """ Request methods """
 
@@ -256,7 +272,7 @@ class AsyncConnection:
                 param = param_line.split('=', 1)
 
                 if len(param) != 2: # Only key received.
-                    log(f'Invalid url path argument', Ansi.RED)
+                    log(f'Invalid url path argument', Ansi.LRED)
                     return
 
                 self.args.update({param[0]: param[1]})
@@ -328,38 +344,40 @@ class AsyncConnection:
         # in our temp buffer. calculate how much we have.
         already_read = len(temp_buf)
 
-        # now that we have our headers, we can allocate our 'real'
-        # buffer with the content length included in the headers.
+        # get the length of the body, and check whether we've
+        # already have the entirety within in our temp buffer.
         content_length = int(self.headers['Content-Length'])
-        buf = bytearray(content_length)
 
-        # add the data we've already read to our
-        # real buffer, and advance the reader.
-        buf[:already_read] = temp_buf
-        view = memoryview(buf)[already_read:]
-        toread = content_length - already_read
+        if already_read != content_length:
+            # there is still data to read; now that we know
+            # the length remaining, we can allocate a static
+            # buffer and read into a view for high efficiency.
+            to_read = content_length - already_read
+            buf = bytearray(to_read)
+            view = memoryview(buf)
 
-        # continue reading from the socket until we've read the
-        # entirety of the data described with the content length.
-        # TODO: implement some sort of timeout system..
-        while toread:
-            nbytes = await loop.sock_recv_into(self.client, view)
-            view = view[nbytes:]
-            toread -= nbytes
+            while to_read:
+                nbytes = await loop.sock_recv_into(self.client, view)
+                view = view[nbytes:]
+                to_read -= nbytes
 
-        # save data to our connection object as immutable bytes.
-        self.body = bytes(buf)
+            # save data to our connection object as immutable bytes.
+            self.body = temp_buf + bytes(buf)
+        else:
+            # we already have all the data.
+            self.body = temp_buf
 
         if self.cmd == 'POST':
             # if we're parsing a POST request, there may
             # still be arguments passed as multipart/form-data.
-            ct = self.headers['Content-Type']
-            if ct and ct.startswith('multipart/form-data'):
-                await self.parse_multipart()
+
+            if 'Content-Type' in self.headers:
+                if self.headers['Content-Type'].startswith('multipart/form-data'):
+                    await self.parse_multipart()
 
     """ Response methods """
 
-    async def add_resp_header(self, header: str, index: int = -1) -> None:
+    def add_resp_header(self, header: str, index: int = -1) -> None:
         if index > -1:
             self.resp_headers.insert(index, header)
         else:
@@ -368,10 +386,10 @@ class AsyncConnection:
     async def send(self, status: Union[HTTPStatus, int], body: bytes = b'') -> None:
         """Attach appropriate headers and send data back to the client."""
         # Insert HTTP response line & content at the beginning of headers.
-        await self.add_resp_header(f'HTTP/1.1 {HTTPStatus(status)!r}'.upper(), 0)
+        self.add_resp_header(f'HTTP/1.1 {HTTPStatus(status)!r}'.upper(), 0)
 
         if body: # Add content-length header if we are sending a body.
-            await self.add_resp_header(f'Content-Length: {len(body)}', 1)
+            self.add_resp_header(f'Content-Length: {len(body)}', 1)
 
         # Encode the headers.
         headers_str = '\r\n'.join(self.resp_headers)
@@ -387,90 +405,97 @@ class AsyncConnection:
         except BrokenPipeError:
             log('Connection ended abruptly', Ansi.LRED)
 
-class AsyncTCPServer:
-    """\
-    Create a TCP socket server which can asynchronously listen and yield connections.
+class Route:
+    """A single endpoint within of domain."""
+    __slots__ = ('methods', 'handler', 'cond')
+    def __init__(self, key: Union[str, Iterable, re.Pattern],
+                 methods: list[str], handler: Callable) -> None:
+        self.methods = methods
+        self.handler = handler
 
-    Simple usage:
-    ```
-    import asyncio
-    import cmyui
-    import time
+        if isinstance(key, str):
+            self.cond = lambda k: k == key
+        elif isinstance(key, Iterable):
+            self.cond = lambda k: k in key
+        elif isinstance(key, re.Pattern):
+            self.cond = lambda k: key.match(k)
 
-    async def handle_conn(conn: cmyui.AsyncConnection) -> None:
-        # see the AsyncConnection implementation for
-        # details on it's use and methods/attributes.
+    def matches(self, key: str, method: str) -> bool:
+        """Check if a given `key` matches our method & condition."""
+        return method in self.methods and self.cond(key)
 
-        # i've provided a simple server example below.
+class RouteMap:
+    """A collection of endpoints of a domain."""
+    __slots__ = ('routes',)
+    def __init__(self) -> None:
+        self.routes = set() # {Route(), ...}
 
-        if 'Host' not in conn.headers:
-            await conn.send(400, b'Missing required headers.')
-            return
+    def route(self, path: Union[str, Iterable[str], re.Pattern],
+              methods: list[str] = ['GET']) -> Callable:
+        """Add a possible route to the server."""
+        if not isinstance(path, (str, Iterable, re.Pattern)):
+            raise TypeError('Route should be str | Iterable[str] | re.Pattern')
 
-        st = time.time()
+        def wrapper(f: Coroutine) -> Coroutine:
+            self.routes.add(Route(path, methods, f))
+            return f
 
-        if conn.cmd == 'GET':
-            if conn.path == '/math/sum.php':
-                if 'x' not in conn.args or 'y' not in conn.args:
-                    await conn.send(400, b'Must supply x & y parameters.')
-                    return
+        return wrapper
 
-                x = conn.args['x']
-                y = conn.args['y']
+    def find_route(self, path: str, method: str):
+        for route in self.routes:
+            if route.matches(path, method):
+                return route
 
-                if not x.isdecimal() or not y.isdecimal():
-                    await conn.send(400, b'Must supply integral parameters.')
-                    return
+class Domain(RouteMap):
+    """The main routemap, with a hostname.
+       Allows for merging of additional routemaps."""
+    __slots__ = ('cond',)
+    def __init__(self, hostname: Union[str, Iterable[str], re.Pattern]) -> None:
+        super().__init__()
 
-                await conn.send(200, f'Sum: {x + y}'.encode())
-                return
-            else:
-                await conn.send(404, b'Handler not found.')
-                return
-        elif conn.cmd == 'POST':
-            if conn.path.startswith('/ss/') and conn.path.endswith('.png'):
-                # POSTing with screenshot in multipart data as a file.
-                if 'screenshot' not in conn.files:
-                    await conn.send(400, b'Missing screenshot data.')
-                    return
-
-                ss_id = conn.path[4:-4]
-
-                if not ss_id.isdecimal():
-                    await conn.send(400, b'Invalid screenshot id.')
-                    return
-
-                with open(f'ss/{ss_id}.png', 'wb') as f:
-                    f.write(conn.files['screenshot'])
-
-                log(f'Saved screenshot {ss_id}.png', Ansi.LGREEN)
-                return
-            else:
-                await conn.send(404, b'Handler not found.')
-                return
+        if isinstance(hostname, str):
+            self.cond = lambda hn: hn == hostname
+        elif isinstance(hostname, Iterable):
+            self.cond = lambda hn: hn in hostname
+        elif isinstance(hostname, re.Pattern):
+            self.cond = lambda hn: hostname.match(hn) is not None
         else:
-            await conn.send(400, b'Handler not found.')
-            return
+            raise TypeError('Key should be str | Iterable[str] | re.Pattern')
 
-    async def run_server():
-        loop = asyncio.get_event_loop()
+    def matches(self, hostname: str) -> bool:
+        """Check if a hostname matches our condition."""
+        return self.cond(hostname)
 
-        # support for both ipv4 and unix domain sockets
-        addr = ('127.0.0.1', 5001) # ipv4
-        addr = '/tmp/gulag.sock' # unix domain
+    def add_map(self, rmap: RouteMap) -> None:
+        self.routes |= rmap.routes
 
-        async with cmyui.AsyncTCPServer(addr) as serv:
-            async for conn in serv.listen(max_conns=16):
-                loop.create_task(handle_conn(conn))
+# TODO: perhaps implement Server, or refactor
+# both into one with a kwarg for async? moyai
+class Server:
+    """An asynchronous multi-domain server."""
+    __slots__ = (
+        'name', 'max_conns', 'gzip',
+        'verbose', 'sock_family',
+        'before_serving', #'after_serving',
+        'domains', 'tasks'
+    )
+    def __init__(self, **kwargs) -> None:
+        self.name = kwargs.get('name', 'Server')
+        self.max_conns = kwargs.get('max_conns', 5)
+        self.gzip = kwargs.get('gzip', 0) # 0-9 valid levels
+        self.verbose = kwargs.get('verbose', False)
+        self.sock_family: Optional[socket.AddressFamily] = None
 
-    asyncio.run(run_server())
-    ```
-    """
-    __slots__ = ('addr', 'sock_family', 'listening')
-    def __init__(self, addr: Address) -> None:
-        is_inet = isinstance(addr, tuple) \
-              and len(addr) == 2 \
-              and all(isinstance(i, t) for i, t in zip(addr, (str, int)))
+        self.before_serving: Optional[Callable] = None
+        #self.after_serving: Optional[Callable] = None # TODO?
+
+        self.domains = set()
+        self.tasks = set()
+
+    def set_sock_mode(self, addr: Address) -> None:
+        is_inet = type(addr) is tuple and len(addr) == 2 and \
+                  type(addr[0]) is str and type(addr[1]) is int
 
         if is_inet:
             self.sock_family = socket.AF_INET
@@ -479,37 +504,157 @@ class AsyncTCPServer:
         else:
             raise ValueError('Invalid address.')
 
-        self.addr = addr
-        self.listening = False
+    # Domain management
 
-    async def __aenter__(self) -> None:
-        return self
+    def add_domain(self, domain: Domain) -> None:
+        self.domains.add(domain)
 
-    async def __aexit__(self, exc_type, exc_val, exc_tb):
-        pass
+    def add_domains(self, domains: set[Domain]) -> None:
+        self.domains |= domains
 
-    async def listen(self, max_conns: int = 5
-                    ) -> AsyncGenerator[AsyncConnection, None]:
-        if using_unix := self.sock_family == socket.AF_UNIX:
-            # Remove unix socket if it already exists.
-            if os.path.exists(self.addr):
-                os.remove(self.addr)
+    def remove_domain(self, domain: Domain) -> None:
+        self.domains.remove(domain)
 
-        loop = asyncio.get_event_loop()
+    def remove_domains(self, domains: set[Domain]) -> None:
+        self.domains -= domains
 
-        sock: socket.socket
-        with socket.socket(self.sock_family, socket.SOCK_STREAM) as sock:
-            sock.bind(self.addr)
+    def find_domain(self, host_header: str):
+        for domain in self.domains:
+            if domain.matches(host_header):
+                return domain
 
-            if using_unix:
-                os.chmod(self.addr, 0o777)
+    # Task management
 
-            self.listening = True
-            sock.listen(max_conns)
-            sock.setblocking(False)
+    def add_task(self, task: asyncio.Task) -> None:
+        self.tasks.add(task)
 
-            while self.listening:
-                client, _ = await loop.sock_accept(sock)
-                conn = AsyncConnection(client)
-                await conn.read()
-                yield conn
+    def remove_task(self, task: asyncio.Task) -> None:
+        self.tasks.remove(task)
+
+    def add_tasks(self, tasks: set[asyncio.Task]) -> None:
+        self.tasks |= tasks
+
+    def remove_tasks(self, tasks: set[asyncio.Task]) -> None:
+        self.tasks -= tasks
+
+    # True Internals
+
+    async def dispatch(self, conn: Connection) -> int:
+        """Dispatch the connection to any matching routes."""
+        host = conn.headers['Host']
+        path = conn.path
+
+        resp = None
+
+        if domain := self.find_domain(host):
+            if route := domain.find_route(path, conn.cmd):
+                resp = await route.handler(conn) or b''
+
+        if resp is not None:
+            code, resp = resp if isinstance(resp, tuple) else (200, resp)
+
+            if self.gzip > 0:
+                resp = gzip.compress(resp, self.gzip)
+                conn.add_resp_header('Content-Encoding: gzip')
+        else:
+            code, resp = (404, b'Not Found.')
+
+        await conn.send(code, resp)
+        return code
+
+    async def handle(self, client: socket.socket) -> None:
+        """Handle a single client socket from the server."""
+        start_time = time.time_ns()
+
+        # Read & parse connection.
+        await (conn := Connection(client)).read()
+
+        if 'Host' not in conn.headers:
+            # This should never happen?
+            client.shutdown(socket.SHUT_RDWR)
+            client.close()
+            return
+
+        # Dispatch the handler.
+        code = await self.dispatch(conn)
+
+        # Event complete, stop timing, log result and cleanup.
+        time_taken = (time.time_ns() - start_time) / 1e6
+
+        colour = (Ansi.LGREEN if 200 <= code < 300 else
+                  Ansi.LYELLOW if 300 <= code < 400 else
+                  Ansi.LRED)
+
+        uri = f'{conn.headers["Host"]}{conn.path}'
+
+        log(f'[{conn.cmd}] {code} {uri}', colour)
+
+        if self.verbose:
+            log(f'Request took {time_taken:.2f}ms', Ansi.LBLUE)
+
+        try:
+            client.shutdown(socket.SHUT_RDWR)
+            client.close()
+        except socket.error:
+            pass
+
+    def run(self, addr: Address) -> None:
+        """Run the server indefinitely."""
+        self.set_sock_mode(addr)
+
+        async def runner() -> None:
+            loop = asyncio.get_event_loop()
+
+            try:
+                loop.add_signal_handler(signal.SIGINT, loop.stop)
+                loop.add_signal_handler(signal.SIGTERM, loop.stop)
+            except NotImplementedError:
+                pass
+
+            # Call our before_serving coroutine,
+            # if theres one specified.
+            if self.before_serving:
+                await self.before_serving()
+
+            # Start up any tasks
+            for task in self.tasks:
+                loop.create_task(task)
+
+            # Setup socket & begin listening
+
+            if self.sock_family == socket.AF_UNIX:
+                if os.path.exists(addr):
+                    os.remove(addr)
+
+            with socket.socket(self.sock_family) as sock:
+                sock.bind(addr)
+
+                if self.sock_family == socket.AF_UNIX:
+                    os.chmod(addr, 0o777)
+
+                sock.listen(self.max_conns)
+                sock.setblocking(False)
+
+                log(f'{self.name} listening @ {addr}', AnsiRGB(0x00ff7f))
+
+                while True:
+                    client, _ = await loop.sock_accept(sock)
+                    loop.create_task(self.handle(client))
+
+        if spec := importlib.util.find_spec('uvloop'):
+            uvloop = importlib.util.module_from_spec(spec)
+            spec.loader.exec_module(uvloop)
+
+            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
+
+        loop = asyncio.new_event_loop()
+
+        try:
+            asyncio.set_event_loop(loop)
+            loop.run_until_complete(runner())
+        finally:
+            loop.run_until_complete(loop.shutdown_asyncgens())
+            # https://github.com/MagicStack/uvloop/pull/353
+            # loop.run_until_complete(loop.shutdown_default_executor())
+            asyncio.set_event_loop(None)
+            loop.close()
